
//
//
// Praktikum 2 in Algorithmen und Datenstrukturen / Fr. Prof. Dr. Kirch-Prinz
// Inhomogene Multilisten
//
// made by
//
// Rene Valdez
// Peter Ellenberg
//
//
// IML.cpp, Implementierung der Methoden
// 

#include "IML.h"
#include <string.h>


////////////////////////////////////////////////
///////////////// class BaseEl /////////////////
////////////////////////////////////////////////
 
// Konstruktor
BaseEl::BaseEl(const string& name, string& telefon) : Cell() {
	setName(name);
	setTel(telefon);
}

// Anzeigen eines Datensatzes
void BaseEl::display() {

	cout << "------- Telefonbucheintrag -------\n" << flush;
	cout << "Name:		" << getName() << "\n" << flush;
	cout << "Telefon:	" << getTel() << "\n" << flush;
}

////////////////////////////////////////////////
///////////////// class ListEl /////////////////
////////////////////////////////////////////////
 
// Konstruktor
ListEl::ListEl(const string& name, string telefon, const string& adresse) : BaseEl(name, telefon) {
	setAdresse(adresse);
}

// Anzeigen eines Datensatzes
void ListEl::display() {

	BaseEl::display();
	cout << "Adresse:	" << flush;
	cout << getAdresse() << "\n" << flush;
}

////////////////////////////////////////////////
/////////////// class InhomList ////////////////
////////////////////////////////////////////////

// Konstruktor
InhomList::InhomList() {

	// Zeiger werden auf NULL initialisiert
	firstName = NULL;
	firstTel = NULL;
	lastName = NULL;
	lastTel = NULL;

	// setzt am Anfang die Anzahl der Elemente auf 0
	anzahl = 0;
}

// Datensatz ohne Adresse eingeben
void InhomList::insert(const string& name, string telefon) {
	insert(name, telefon, "");
}

void InhomList::insert(const string& name, string telefon, const string& adresse) {

	// Zeiger auf eine Zelle initialisieren
	Cell* cellName		= NULL;
	Cell* cellTel		= NULL;
	Cell* nextCellName	= NULL;
	Cell* nextCellTel	= NULL;
	Cell* temp			= NULL;
	BaseEl* baseEl;

	// Wenn eine Adresse dabei ist
	if ( adresse != "" ) {

		// wird ein Element des Typs ListEl erstellt
		ListEl* listEl = new ListEl(name, telefon, adresse);

		// Es wird allerdings mit dem gecasteten Zeiger gearbeitet !
		baseEl = (BaseEl*)listEl;

	} else {

		// ansonsten ein Element des Typs BaseEl
		baseEl = new BaseEl(name, telefon);
	}
	
	// Wenn noch kein Element in der Liste ist
	if ( getAnzahl() == 0 ) {
		insertName(NULL, NULL, (Cell*)baseEl);
		insertTel(NULL, NULL, (Cell*)baseEl);
		setAnzahl(1);
	} else {

		// Flag, um die Laufzeit zu verkrzen
		// wenn flag = 2, dann wurde der Name und die Telefonnummer verlinkt
		int flag = 0;

		// zuallererst die Sonderflle betrachten:
		// Der Name des neuen Elements ist kleiner als der Name des Ersten
		if ( dynamic_cast<BaseEl*>(getFirstName())->getName() > baseEl->getName() ) {
			insertName(NULL, getFirstName(), (Cell*)baseEl);
			flag++;
		}

		// Der Telefonnummer des neuen Elements ist kleiner als die Telefonnummer des Ersten
		if ( dynamic_cast<BaseEl*>(getFirstTel())->getTel() > baseEl->getTel() ) {
			insertTel(NULL, getFirstTel(), (Cell*)baseEl);
			flag++;
		}

		// Der Name des neuen Elements ist grer als der Name des Letzten
		if ( dynamic_cast<BaseEl*>(getLastName())->getName() < baseEl->getName() ) {
			insertName(getLastName(), NULL, (Cell*)baseEl);
			flag++;
		}

		// Der Telefonnummer des neuen Elements ist grer als die Telefonnummer des Letzten
		if ( dynamic_cast<BaseEl*>(getLastTel())->getTel() < baseEl->getTel() ) {
			insertTel(getLastTel(), NULL, (Cell*)baseEl);
			flag++;
		}

		// Zeiger auf das erste Element schalten
		cellName	= getFirstName();
		cellTel		= getFirstTel();

		// Nachfolger des ersten Elements 
		// (der Vorgnger des ersten Elements ist NULL)
		nextCellName	= cellName->getNextName(NULL);
		nextCellTel		= cellTel->getNextTel(NULL);

		// geht die ganze Liste durch
		for ( int i=0; i < getAnzahl(); i++ ) {

			// Es wurden schon beide Links gesetzt
			// -> Schleife kann vorzeitig beendet werden
			if ( flag == 2) {
				setAnzahl(getAnzahl()+1);
				break;
			}

			// der Link fr den Namen wurde noch nicht gesetzt
			if ( baseEl->getLink1() == NULL ) {

				// wenn der folgende Eintrag grer ist als der jetzige
				if ( dynamic_cast<BaseEl*>(nextCellName)->getName() > baseEl->getName() ) {
					insertName(cellName, nextCellName, (Cell*)baseEl);
					flag++;
				}
			}

			// der Link fr die Telefonnummer wurde noch nicht gesetzt
			if ( baseEl->getLink2() == NULL ) {

				// wenn der folgende Eintrag grer ist als der jetzige
				if ( dynamic_cast<BaseEl*>(nextCellTel)->getTel() > baseEl->getTel() ) {
					insertTel(cellTel, nextCellTel, (Cell*)baseEl);
					flag++;
				}
			}

			// Hilfsvariable zum Vertauschen
			temp = cellName;
			cellName = nextCellName;
			nextCellName = nextCellName->getNextName(temp);
			temp = cellTel;
			cellTel = nextCellTel;
			nextCellTel = nextCellTel->getNextTel(temp);

		}
	}
}

// Lscht eine Zelle aus der Liste
void InhomList::deleteCell(const string& name) {

	// Pointer auf Pointerarray von Cell, in dem die Pointer auf die Vorgnger- ([0]), 
	// die gesuchte ([1]), und die Nachfolgerzelle ([2]) sind
	Cell** foundCellName;
	Cell** foundCellTel;

	// sucht nach dem Cell Pointer dieses Namens und dessen Vorgnger und Nachfolger
	foundCellName = findCellName(name);

	if (foundCellName == NULL) {
		cout << "Eintrag "<< name << " konnte nicht gefunden werden !\n" << flush;
		return;
	}

	if ( getAnzahl()==1 ) {

		// Anzahl wird auf 0 gesetzt
		setAnzahl(0);

		// Es gibt keine Anfnge mehr
		setFirstName(NULL);
		setFirstTel(NULL);
		setLastName(NULL);
		setLastTel(NULL);

		// das noch brige Element wird gelscht
		delete foundCellName[1];
		return;
	}

	// sucht nach dem Cell Pointer dieser Telefonnummer, welcher natrlich der 
	// gleiche ist wie fr den Namen. Wir interessieren uns aber fr 
	// dessen Vorgnger und Nachfolger, die natrlich nicht die gleichen sein mssen, 
	// wie fr den Namen !
	foundCellTel = findCellTel(dynamic_cast<BaseEl*>(foundCellName[1])->getTel());

	// Wenn ein Sonderfall auftritt:
	if (foundCellName[0] == NULL || foundCellName[2] == NULL) {

		// Sonderflle
		if (foundCellName[0] == NULL) {
			setFirstName(foundCellName[2]);
			foundCellName[2]->setLink1(NULL, foundCellName[2]->getNextName(foundCellName[1]));
		}
		if (foundCellName[2] == NULL) {
			setLastName(foundCellName[0]);
			foundCellName[0]->setLink1(foundCellName[0]->getNextName(foundCellName[1]), NULL);
		}
	} else {
		foundCellName[0]->setLink1(foundCellName[0]->getNextName(foundCellName[1]), foundCellName[2]);
		foundCellName[2]->setLink1(foundCellName[0], foundCellName[2]->getNextName(foundCellName[1]));
	}

	// Wenn ein Sonderfall auftritt:
	if (foundCellTel[0] == NULL || foundCellTel[2] == NULL) {
		if (foundCellTel[0] == NULL) {
			setFirstTel(foundCellTel[2]);
			foundCellTel[2]->setLink2(NULL, foundCellTel[2]->getNextTel(foundCellTel[1]));
		}

		if (foundCellTel[2] == NULL) {
			setLastTel(foundCellTel[0]);
			foundCellTel[0]->setLink2(foundCellTel[0]->getNextTel(foundCellTel[1]), NULL);
		}
	} else {
		foundCellTel[0]->setLink2(foundCellTel[0]->getNextTel(foundCellTel[1]), foundCellTel[2]);
		foundCellTel[2]->setLink2(foundCellTel[0], foundCellTel[2]->getNextTel(foundCellTel[1]));
	}

	// Objekt wird gelscht
	delete foundCellName[1];

	// die Anzahl wird verringert
	setAnzahl(getAnzahl()-1);
}

// gibt fr einen bestimmten Namen die Zelle mit Vorgnger und Nachfolger zurck
Cell** InhomList::findCellName(const string& name) {

	Cell* cell;
	Cell* nextCell;
	Cell* temp;
	Cell** findCell = new Cell*[3];
	
	cell = NULL;
	nextCell = firstName;
	
	// durchluft die Liste
	while (nextCell != NULL) {
		if ( dynamic_cast<BaseEl*>(nextCell)->getName() == name ) {

			// Vorgnger
			findCell[0] = cell;

			// gefundene Zelle
			findCell[1] = nextCell;

			// Nachfolger
			findCell[2] = nextCell->getNextName(cell);

			return findCell;
		}

		temp = cell;
		cell = nextCell;
		nextCell = nextCell->getNextName(temp);
	}

	// Element wurde nicht gefunden !
	return NULL;

}

// gibt fr eine bestimmte Telefonnummer die Zelle mit Vorgnger und Nachfolger zurck
Cell** InhomList::findCellTel(string telefon) {

	Cell* cell;
	Cell* nextCell;
	Cell* temp;
	Cell** findCell = new Cell*[3];
	cell = NULL;
	nextCell = firstTel;
	
	// durchluft die Liste
	while (nextCell != NULL) {

		if ( dynamic_cast<BaseEl*>(nextCell)->getTel() == telefon ) {

			// Vorgnger
			findCell[0] = cell;

			// gefundene Zelle
			findCell[1] = nextCell;

			// Nachfolger
			findCell[2] = nextCell->getNextTel(cell);
			return findCell;
		}
		temp = cell;
		cell = nextCell;
		nextCell = nextCell->getNextTel(temp);
	}

	// Element wurde nicht gefunden !
	return NULL;
}

// setzt den neuen Link fr den Namen
void InhomList::insertName(Cell* cellName, Cell* nextCellName, Cell* newCell) {
	
	// Es tritt ein Sonderfall ein
	if (cellName == NULL || nextCellName == NULL) {

		// Sonderfall firstName()
		if (cellName == NULL && nextCellName != NULL) {

			// den Link des Vorgngers verbiegen
			getFirstName()->setLink1(newCell, getFirstName()->getNextName(NULL));

			// den eigenen Link verrechnen mit vorherigen Wert und NULL
			newCell->setLink1(getFirstName(), NULL);

			// den Pointer auf das aktuelle Element setzen
			setFirstName(newCell);

		}

		// Sonderfall lastName()
		if (cellName != NULL && nextCellName == NULL) {
			
			// den Link des Vorgngers verbiegen
			getLastName()->setLink1(getLastName()->getNextName(NULL), newCell);

			// den eigenen Link verrechnen mit vorherigen Wert und NULL
			newCell->setLink1(getLastName(), NULL);

			// den Pointer auf das aktuelle Element setzen
			setLastName(newCell);
		}

		// Sonderfall es gibt noch kein Element
		if (cellName == NULL && nextCellName == NULL) {

			// setzt nun die beiden Links
			// (vorn und hinten ist nix !)
			newCell->setLink1(NULL, NULL);


			// setzt die Pointer auf das erste Element
			setFirstName(newCell);
			setLastName(newCell);
		}
	}
	else {

		// dabei werden der Vorgnger der Zelle mit dem neuen Element verknpft
		cellName->setLink1(cellName->getNextName(nextCellName), newCell);

		// neue Zelle mit dem Vorgnger und Nachfolger verrechnen
		newCell->setLink1(cellName, nextCellName);

		// anschlieend wird die Kette wieder geschlossen
		// indem die Nachfolger Zelle mit der neuen Zelle verknpft werden
		nextCellName->setLink1(newCell, nextCellName->getNextName(cellName));

	}
}

// setzt den neuen Link fr die Telefonnummer
void InhomList::insertTel(Cell* cellTel, Cell* nextCellTel, Cell* newCell) {
	
	// Es tritt ein Sonderfall ein
	if (cellTel == NULL || nextCellTel == NULL) {

		// Sonderfall firstTel()
		if (cellTel == NULL && nextCellTel != NULL) {

			// den Link des Vorgngers verbiegen
			getFirstTel()->setLink2(newCell, getFirstTel()->getNextTel(NULL));

			// den eigenen Link verrechnen mit vorherigen Wert und NULL
			newCell->setLink2(getFirstTel(), NULL);

			// den Pointer auf das aktuelle Element setzen
			setFirstTel(newCell);

		}

		// Sonderfall lastTel()
		if (cellTel != NULL && nextCellTel == NULL) {
			
			// den Link des Vorgngers verbiegen
			getLastTel()->setLink2(getLastTel()->getNextTel(NULL), newCell);

			// den eigenen Link verrechnen mit vorherigen Wert und NULL
			newCell->setLink2(getLastTel(), NULL);

			// den Pointer auf das aktuelle Element setzen
			setLastTel(newCell);
		}

		// Sonderfall es gibt noch kein Element
		if (cellTel == NULL && nextCellTel == NULL) {

			// setzt nun die beiden Links
			// (vorn und hinten ist nix !)
			newCell->setLink2(NULL, NULL);

			// setzt die Pointer auf das erste Element
			setFirstTel(newCell);
			setLastTel(newCell);
		}
	}
	else {

		// dabei werden der Vorgnger der Zelle mit dem neuen Element verknpft
		cellTel->setLink2(cellTel->getNextTel(nextCellTel), newCell);

		// neue Zelle mit dem Vorgnger und Nachfolger verrechnen
		newCell->setLink2(cellTel, nextCellTel);

		// anschlieend wird die Kette wieder geschlossen
		// indem die Nachfolger Zelle mit der neuen Zelle verknpft werden
		nextCellTel->setLink2(newCell, nextCellTel->getNextTel(cellTel));
	}
}

// berladene Methode, um nach einen Namen zu suchen
void InhomList::retrieve(const string& name, int was) {

	if (was == 0)
	{

		Cell** foundCellName;
		// sucht nach dem Element
		foundCellName = findCellName(name);
		if (foundCellName == NULL) {
			cout << "Fehler !\nName "<< name <<" wurde in Liste nicht gefunden !\n" << flush;
			return;
		}
		foundCellName[1]->display();

	} else {
		
		Cell** foundCellTel;
		// sucht nach dem Element
		foundCellTel = findCellTel(name);
		if (foundCellTel == NULL) {
			cout << "Fehler !\nTelefonnnummer "<< name <<" wurde in Liste nicht gefunden !\n" << flush;
			return;
		}
		foundCellTel[1]->display();
	}

}

// geht die Liste sortiert nach Namen oder Telefon von vorne oder von hinten durch und zeigt diese an
void InhomList::traverse(int was, int vonWo) {
	
	Cell* cell;
	Cell* nextCell;
	BaseEl* baseEl;
	ListEl* listEl;

	if (getAnzahl()==0) {
		cout << "Die Liste hat keinen Eintrag !\n" << flush;
		return;
	}

	cell = NULL;

	// Hier wird unterschieden, was zu tun ist
	if (was == LEXIKON && vonWo == AUFSTEIGEND) nextCell = firstName;
	if (was == LEXIKON && vonWo == ABSTEIGEND) nextCell = lastName;
	if (was == NUMMER && vonWo == AUFSTEIGEND) nextCell = firstTel;
	if (was == NUMMER && vonWo == ABSTEIGEND) nextCell = lastTel;

	while ( nextCell != NULL ) {

		baseEl = dynamic_cast<BaseEl*>(nextCell);
		listEl = dynamic_cast<ListEl*>(nextCell);

		if ( baseEl != NULL ) {
			baseEl->display();
			cout << endl;
		} else if ( listEl != NULL ) {
			listEl->display();
			cout << endl;
		} else cout << "\n dynamic_cast<> Error ! \n" << flush; 

		Cell* temp;
		temp = cell;
		cell = nextCell;
		if (was == LEXIKON)	nextCell = nextCell->getNextName(temp);
		else nextCell = nextCell->getNextTel(temp);
	}
}

// Destruktor baut die ganze Liste ab und gibt den Speicher wieder frei
InhomList::~InhomList() {

	// hier werden alle Elemente gelscht
	while ( getAnzahl() != 0 ) {
		deleteCell(dynamic_cast<BaseEl*>(getFirstName())->getName());
	}

}